#!/usr/bin/env python
import re
import sys
import subprocess
import os
import csv
import time
import select

re1 = re.compile(r"For Correlation Point:[^\n]*((?:.|\n)*?)\{\s*char \*CP([0-9]+);\s*\}", re.MULTILINE)
re2 = re.compile(r"State = \[\s*((?:.|\n)*)\s*\]", re.MULTILINE)
re3 = re.compile(r"=\(([a-zA-Z0-9_, ]*)\)\{[^}]*\}[^<]*<-> =\(([a-zA-Z0-9_, ]*)\)", re.MULTILINE)

RES_TIMEOUT, RES_EQUIVALENT, RES_DIFFERENT = range(3)

def main(path):
    with open(path, 'r') as in_file:
        reader = csv.reader(in_file)
        results = []
        for (p, expected, args) in reader:
            res = analyze(p, args.split(' ') if args else None)
            n = res[0]
            stuff = res[1:]
            results.append((n, expected,) + stuff)

    with open('results.csv', 'w') as out_file:
        writer = csv.writer(out_file)
        writer.writerow(('file', 'expected', 'hascounter',
                         'cccresppl', 'ccctimeppl',
                         'cpresppl', 'cptimeppl'))
        for r in results:
            writer.writerow(r)

def analyze(path, args):
    print('Analyzing %s...' % path)
    dirname = os.path.dirname(path)
    basename = os.path.basename(path)
    filename, ext = os.path.splitext(basename)
    del path
    del basename

    if ext == '.imp':
        build_diffmap(dirname, filename)
        ext = '.diffmap'

    build_union_cp(dirname, filename)
    pygmentize(dirname, filename + '.imp', filename + '.imp.html', 'c')
    pygmentize(dirname, 'patched.' + filename + '.imp', 'patched.' + filename + '.imp.html', 'c')

    to_text = {RES_TIMEOUT: 'timeout', RES_EQUIVALENT: 'equivalent', RES_DIFFERENT: 'nonequivalent'}.get

    results_cp = []
    for m in ('ppl',):
        time_cp, equivalent_cp = interpret_dizy_output(dirname, filename, domain=m)
        results_cp.append(('%.02f' % time_cp, to_text(equivalent_cp)))

    build_union_cp(dirname, filename, insert_cps=True)
#    results_cpcp = []
#    for m in ('oct', 'ppl'):
#        time_cpcp, equivalent_cpcp = interpret_dizy_output(dirname, filename, domain=m)
#        results_cpcp.append(('%.02f' % time_cpcp, to_text(equivalent_cpcp)))

    build_union_ccc(dirname, filename)
    results_ccc = []
    for m in ('ppl',):
        time_ccc, equivalent_ccc = interpret_dizy_output(dirname, filename, prefix='', domain=m)
        results_ccc.append(('%.02f' % time_ccc, to_text(equivalent_ccc)))

    if args:
        build_mains(dirname, filename)
        ccc_counter_example = check_counter_example(dirname, filename, args)
    else:
        ccc_counter_example = False

    return (filename, ccc_counter_example,
                      results_ccc[0][1], results_ccc[0][0],
#                      results_ccc[1][1], results_ccc[1][0],
                      results_cp[0][1], results_cp[0][0])
#                      results_cp[1][1], results_cp[1][0],
#                      results_cpcp[0][1], results_cpcp[0][0],
#                      results_cpcp[1][1], results_cpcp[1][0])


def build_diffmap(dirname, filename):
    subprocess.check_call(['syntacticdiff.native',
        os.path.join(dirname, filename + '.imp'),
        os.path.join(dirname, 'patched.' + filename + '.imp'),
        os.path.join(dirname, filename + '.diffmap')])
    pygmentize(dirname, filename + '.diffmap', filename + '.diffmap.html', 'diff')


def build_union_ccc(dirname, filename, only_last_cp=True):
    ccc_union_path = os.path.join(dirname, 'union.' + filename + '.c')
    subprocess.check_call(['union.sh',
        os.path.join(dirname, filename + '.c'),
        os.path.join(dirname, 'patched.' + filename + '.c')])
    subprocess.check_call(['sed', '-i', 's/#include.*//',
        os.path.join(dirname, 'union.' + filename + '.c')])
    subprocess.check_call(['sed', '-i', 's/typedef int _Bool;//',
        ccc_union_path])

    if only_last_cp:
        regexp = re.compile(r"\s*\{char \*CP[0-9]+\;\}")
        last_cp = (-1, '')
        with open(ccc_union_path, 'r') as file:
            lines = []
            for i, l in enumerate(file):
                if regexp.match(l):
                    lines.append('/* ' + l.rstrip('\n') + ' */\n')
                    last_cp = (i, l)
                else:
                    lines.append(l)
        if last_cp[0] >= 0:
            lines[last_cp[0]] = last_cp[1]
        with open(ccc_union_path, 'w') as file:
            for l in lines:
                file.write(l)
    pygmentize(dirname, 'union.' + filename + '.c', 'union.' + filename + '.ccc.html', 'c')


def check_counter_example(dirname, filename, args):
    dirpath = os.path.abspath(dirname)
    orig = subprocess.check_output([os.path.join(dirpath, filename)] + args)
    patched = subprocess.check_output([os.path.join(dirpath, 'patched.' + filename)] + args)
    union = subprocess.check_output([os.path.join(dirpath, 'union.' + filename)] + args)
    return [orig.strip(), patched.strip()] != [x.strip() for x in union.split('\n')]


def build_mains(dirname, filename):
    subprocess.check_call(['gcc', '-o',
        os.path.join(dirname, filename),
        os.path.join(dirname, filename + '.c'),
        os.path.join(dirname, 'main.' + filename + '.c')])
    subprocess.check_call(['gcc', '-o',
        os.path.join(dirname, 'patched.' + filename),
        os.path.join(dirname, 'patched.' + filename + '.c'),
        os.path.join(dirname, 'main.patched.' + filename + '.c')])
    subprocess.check_call(['gcc', '-o',
        os.path.join(dirname, 'union.' + filename),
        os.path.join(dirname, 'union.' + filename + '.c'),
        os.path.join(dirname, 'main.union.' + filename + '.c')])


def build_union_cp(dirname, filename, insert_cps=False):
    options = ['--enable-projs', '--enable-main']
    if insert_cps:
        options.append('--insert-cps')
    subprocess.check_call(['correlating_program.native'] + options +
                          [os.path.join(dirname, filename + '.diffmap')])
    pygmentize(dirname, 'union.' + filename + '.c',
               'union.' + filename + ('.cp2.html' if insert_cps else '.cp.html'), 'c')


def pygmentize(dirname, infile, outfile, lexer):
    subprocess.check_call(['pygmentize', '-O',
        'full,linenos=1,cssfile=pygments.css',
        '-o', outfile,
        '-l', lexer,
        infile])


def interpret_dizy_output(dirname, filename, prefix='O_', domain='ppl', timeout=1200):
    p = subprocess.Popen(['dizy', '-w_s=guard', '-m=' + domain,
        os.path.join(dirname, 'union.' + filename + '.c')],
        stderr=subprocess.STDOUT, stdout=subprocess.PIPE,
        bufsize=1)

    c = ''

    # Handle timeout:
    start = time.time()
    while time.time() - start < timeout and p.poll() is None:
        time.sleep(0.01)

    if p.poll() is None:
        p.kill()
        print('PROCESS KILLED (timeout)')
        return (timeout, RES_TIMEOUT)
    else:
        time_ellapsed = time.time() - start

    c = p.stdout.read()

    last_state_ok = False
    ok = False
    for (body, cpnum) in re1.findall(c):
        last_state_ok = False
        cpnum = int(cpnum)
        state = re2.findall(body)[0].strip()
        #XXX: rewrite read(x, y) and update(x, y)
        state = re.sub(r"(read|update)\([a-zA-Z0-9_, ]*\)", "", state)
        vars_preserved = None
        print('Correlation Point CP%d:' % cpnum)
        for i, (guards, vars) in enumerate(re3.findall(state)):
            guards = filter(None, guards.split(','))
            vars = filter(None, vars.split(','))
            if vars_preserved is None:
                vars_preserved = set(vars)
            vars_preserved.intersection_update(vars)
            print('  Preserved variables in sub-state #%d: %s' % (i, ', '.join(sorted(x[len(prefix):] for x in vars)) if vars else '(none)'))

        print('  Variables preserved in all sub-states: %s' % (', '.join(sorted(x[len(prefix):] for x in vars_preserved)) if vars_preserved else '(none)'))

        if vars_preserved and (prefix + 'result') in vars_preserved:
            last_state_ok = True

    if last_state_ok:
        print('Result value preserved!')
        return (time_ellapsed, RES_EQUIVALENT)
    else:
        print('Result value MIGHT BE DIFFERENT IN THE TWO PROGRAMS!')
        return (time_ellapsed, RES_DIFFERENT)


if __name__ == '__main__':
  main(*sys.argv[1:])
